/*
 *  SerialServer.cpp
 *  polynomeSerial
 *
 *  Created by Colin Mann on 5/5/10.
 *  Copyright 2010 Mannchester Analog. All rights reserved.
 *
 */
using namespace std;

#include "SerialServer.h"
#include <termios.h>
#include <sys/ioctl.h>
#include <iostream>
#include <dirent.h>
#include <string>
#include <fcntl.h>


void * theServer(void *context) {
	
	
	//Setup serial port nfs select stuff here.
	int nfds = 0;
	fd_set readfds;
	struct timeval timeout;
	//size_t bytesToRead;
	size_t bytesRead;
	unsigned char buffer[2];
	memset(buffer,0,2);
	
	while(1) {
		nfds = 0;
		FD_ZERO(&readfds);
		
		fd = context->fileDescriptor;
		
		if(fd >= nfds)
			nfds = fd+1;
		
		FD_SET(fd, &readfds);
		
		timeout.tv_sec = 1;
		timeout.tv_usec = 0;
		
		nfds = select(nfds, &readfds, NULL, NULL, &timeout);

		if(nfds > 0)
		{
			bytesRead = read(fd, buffer,2);
			while(bytesRead<2)
				bytesRead += read(fd,buffer,1);
		}
	}
	pthread_exit(NULL);
	return NULL;
	
}

SerialServer::SerialServer() {
	bVerbose = false;
}

SerialServer::SerialServer(string device_name, speed_t baud, size_t messageSize){
	
	string device;
	DIR *dir;
	struct dirent *entries;
	dir = opendir("/dev");
	
	if(dir == NULL) {
		cout<<"Error listing devices in /dev\n";
	}
	
	while ((entries = readdir(dir)) != NULL) {
		device = (char *)entries->d_name;
		if( (device.rfind(device_name)!= (string::npos)) ) {
			fd = open(("/dev/"+device).c_str(), O_RDWR | O_NOCTTY | O_NONBLOCK);
			break;
		}
	}
	
	if(fd == -1)
		cout<<"Unable to open port.\n";
	else {
		struct termios options;
		tcgetattr(fd,&oldoptions);
		options = oldoptions;
		
		cfsetispeed(&options,BAUD_RATE);
		cfsetospeed(&options,BAUD_RATE);
		
		options.c_cflag |= (CLOCAL | CREAD);
		options.c_cflag &= ~PARENB;
		options.c_cflag &= ~CSTOPB;
		options.c_cflag &= ~CSIZE;
		options.c_cflag |= CS8;
		tcsetattr(fd,TCSANOW,&options);
		
		bVerbose = false;
		startSerialServer(fd, messageSize);
	}
}

SerialServer::~SerialServer() {
	
	close();
	bVerbose = false;
}

void SerialServer::startSerialServer(fd_set thisdevice, size_t pSize) {
	
	if(serial_thread != 0)
		return;

	SerialDeviceContext context = { thisdevice, pSize };
	
	pthread_create(&serial_thread, NULL, theServer, (void *)&context);
	
}


void SerialServer::setVerbose(bool b) {
	bVerbose = b;
}

void SerialServer::close() {
		tcsetattr(fd,TCSANOW,&oldoptions);
		::close(fd);
}


int SerialServer::writeBytes(unsigned char * buffer, int length) {
	int byteWritten= write(fd, buffer, length);
	if(byteWritten <= 0) {
		cout<<"Error writing to serial port.\n";
	}
	return byteWritten;
}


int SerialServer::writeByte(unsigned char b) {
	unsigned char byte[1];
	byte[0] = b;
	int byteWritten= write(fd, byte, 1);
	if(byteWritten <= 0) {
		cout<<"Error writing to serial port.\n";
	}
	return byteWritten;
}


void SerialServer::flush(bool flushIn, bool flushOut) {
	int flushType = 0;
	
	if( flushIn && flushOut) flushType = TCIOFLUSH;
	else if(flushIn) flushType = TCIFLUSH;
	else if(flushOut) flushType = TCOFLUSH;
	else return;
	
	tcflush(fd, flushType);
}

int SerialServer::available() {

	
	int num=0;
	ioctl(fd, FIONREAD,&num);
	return num;
}

