/*
 *  polynome.cpp
 *  polynomeSerial
 *
 *  Created by Colin Mann on 5/26/10.
 *  Copyright 2010 Mannchester Analog. All rights reserved.
 *
 */

#include <stdio.h>
#include <iostream>
#include <stdint.h>
#include "lo/lo.h"
#include "SerialDevice.h"
#include "polynome_message.h"
#include "polynome.h"

typedef uint8_t		uint8;


using namespace std;

static int _Polynome_SerialDeviceMessageReceivedCallback(SerialDevice *device, char *data, size_t len, void *userData)
{
	polynome *SELF = (polynome *)userData;
	p_message *message = (p_message *)data;
	string keypadConvert = "123456789*0#";
	int bcd=0;
	switch (messageGetType(*message)) {
		case M_KEYDOWN:
			lo_send(SELF->getMonomeSender(),(SELF->getPrefix()+"/press").c_str(),"iii",(message->data1)>>4,(message->data1)&0x0F,1);
			break;
		case M_KEYUP:
			lo_send(SELF->getMonomeSender(),(SELF->getPrefix()+"/press").c_str(),"iii",(message->data1)>>4,(message->data1)&0x0F,0);
			break;
		case ANALOGIN:
			if ( ((message->data0)&0x0F)==6) {
				vector<unsigned char> midimessage;
				midimessage.push_back(PMIDI_CC_MESSAGE);
				midimessage.push_back( ((message->data0)&0x0F)+PMIDI_ANALOG_STARTCC);
				int ir = (message->data1)*5/3;
				midimessage.push_back(ir);
				SELF->getMidiOut()->sendMessage( &midimessage );
				lo_send(SELF->getPolynomeSender(),"/polynome/fader","ii",message->data0-32,ir);
			}
			else if ( ((message->data0)&0x0F)==7) {
				vector<unsigned char> midimessage;
				midimessage.push_back(PMIDI_CC_MESSAGE);
				int slide = (message->data1)*1.02;
				midimessage.push_back( ((message->data0)&0x0F)+PMIDI_ANALOG_STARTCC);
				midimessage.push_back(slide);
				SELF->getMidiOut()->sendMessage( &midimessage );
				lo_send(SELF->getPolynomeSender(),"/polynome/fader","ii",message->data0-32,slide);
			}
			else {
				vector<unsigned char> midimessage;
				midimessage.push_back(PMIDI_CC_MESSAGE);
				midimessage.push_back( ((message->data0)&0x0F)+PMIDI_ANALOG_STARTCC);
				midimessage.push_back(message->data1);
				SELF->getMidiOut()->sendMessage( &midimessage );
				lo_send(SELF->getPolynomeSender(),"/polynome/fader","ii",message->data0-32,message->data1);
			}
			break;
		case BUTTONDOWN:
			lo_send(SELF->getPolynomeSender(),"/polynome/arcade","ii",(message->data1),1);
			printf("POLYNOME:ARCADE:%i\t\ton\n",message->data1);
			break;
		case BUTTONUP:
			lo_send(SELF->getPolynomeSender(),"/polynome/arcade","ii",(message->data1),0);
			printf("POLYNOME:ARCADE:%i\t\toff\n",message->data1);
			break;
		case KEYPAD:
			bcd = SELF->keypadAnalyze((keypadConvert[(message->data1)&0x0F]));
			uint8 m[4];
			m[0] = NIXIESET;
			m[1] = bcd>>8;
			m[2] = bcd & 0x00FF;
			m[3] = NIXIEDEC;
			SELF->getSerialDevice()->write((char *)m, 4);
			break;
		case KEYPADUP:
			break;
		case MODE:
			SELF->setMode(message->data1);
			lo_send(SELF->getPolynomeSender(),"/polynome/mode","i",message->data1);
			printf("POLYNOME:MODE:%i\n",message->data1);
			break;
		case DEBUG:
			switch (message->data1) {
				case UART_FRAME_ERROR:
					printf("POLYNOME:A framing error has occured on the UART bus.\n");
					break;
				case UART_OVERRUN_ERROR:
					printf("POLYNOME:A buffer overrun error has occured, sending data too fast.\n");
					break;
				case UART_BUFFER_OVERFLOW:
					printf("POLYNOME:A buffer overflow error has occured, sending data too fast.\n");
					break;
				default:
					printf("POLYNOME:DEBUG:DEFAULT: byte 0x%X\n",message->data1);
			}
			break;
		default:
			printf("POLYNOME:DEFAULT:Byte 0 is: 0x%X, Byte 1 is: 0x%X\n",(message->data0),message->data1);
			return -1;
			break;
	}
	return 0;
	
}

void polynomeMidiCallback( double deltatime, std::vector< unsigned char > *message, void *userData )
{
	unsigned int nBytes = message->size();
	for ( unsigned int i=0; i<nBytes; i++ )
		cout << "Byte " << i << " = " << (int)message->at(i) << ", ";
	if ( nBytes > 0 )
		cout << "stamp = " << deltatime << endl;
}


polynome::polynome() {
	
	_prefix = "/40h";
	
	_nixieUpdate = true;
	_nixieDisplay = "0000";
	_nixieBCD = 0;
	
	clearColorMatrix();
	
}


polynome::~polynome() {
	_polyDeviceReader.stopReading();
	_polyDeviceReader.removeSerialDevice(_polySerial);
	_polyMidiOut->closePort();
	delete _polyMidiOut;
	delete _polyMidiIn;
}


void polynome::start() {
	
	_polyMidiOut = new RtMidiOut();
	_polyMidiOut->openVirtualPort("from polynome");
	
	_polySerial = new SerialDevice("/dev/tty.usbserial-polynome", B230400);
	_polyDeviceReader.addSerialDevice(_polySerial, 2, _Polynome_SerialDeviceMessageReceivedCallback, this);
	_polySerial->flush();
	
	//Create osc thread
	_polyOscThread = lo_server_thread_new(POLYNOME_LISTEN_PORT, polyOscError);
	_monomeOscThread = lo_server_thread_new(MONOME_LISTEN_PORT, polyOscError);
	//Create osc message senders
	_polynomeOscSender = lo_address_new(NULL, POLYOSC_PORT_SENDER);	//Null address is localmachine
	_monomeOscSender = lo_address_new(NULL, MONOMEOSC_PORT_SENDER);
	
	//Monome server
	lo_server_thread_add_method(_monomeOscThread, (_prefix+"/led").c_str(), "iii", m_LedHandler, this);
	lo_server_thread_add_method(_monomeOscThread, (_prefix+"/clear").c_str(), "i", m_ClearHandler, this);
	lo_server_thread_add_method(_monomeOscThread, (_prefix+"/led_col").c_str(), "ii", m_LedColHandler, this);
	lo_server_thread_add_method(_monomeOscThread, (_prefix+"/led_row").c_str(), "ii", m_LedRowHandler, this);
	lo_server_thread_add_method(_monomeOscThread, (_prefix+"/frame").c_str(), "iiiiiiii", m_frameHandler, this);
	lo_server_thread_add_method(_monomeOscThread, "/sys/prefix", "s", m_SysPrefixHandler, this);
	lo_server_thread_add_method(_monomeOscThread, "/sys/test", "i", m_SysPrefixHandler, this);
	lo_server_thread_add_method(_monomeOscThread, "/sys/prefix", "is", m_SysPrefixHandler, this);
	lo_server_thread_add_method(_monomeOscThread, "/sys/report", "", m_SysReportHandler, this);
	
	//Polynome Server
	lo_server_thread_add_method(_polyOscThread, "/polynome/mode", "i", p_ModeHandler, this);
	lo_server_thread_add_method(_polyOscThread, "/polynome/monocolor", "iii", p_MonoColorHandler, this);
	lo_server_thread_add_method(_polyOscThread, "/polynome/led", "iiiii", p_LedHandler, this);
	lo_server_thread_add_method(_polyOscThread, "/polynome/row", "iiii", p_RowHandler, this);
	lo_server_thread_add_method(_polyOscThread, "/polynome/col", "iiii", p_ColHandler, this);
	lo_server_thread_add_method(_polyOscThread, "/polynome/colordump", "s", p_ColorDumpHandler, this);
	lo_server_thread_add_method(_polyOscThread, "/polynome/colordump", "siiii", p_ColorDumpHandler, this);
	lo_server_thread_add_method(_polyOscThread, "/polynome/colordump", "siiiii", p_ColorDumpHandler, this);
	lo_server_thread_add_method(_polyOscThread, "/polynome/arcade", "ii", p_ArcadeHandler, this);
	lo_server_thread_add_method(_polyOscThread, "/polynome/nixie", "i", p_NixieHandler, this);
	lo_server_thread_add_method(_polyOscThread, "/polynome/nixie", "f", p_NixieHandler, this);
	lo_server_thread_add_method(_polyOscThread, "/polynome/update", "i", p_UpdateHandler, this);
	
	_polyDeviceReader.startReading();
	lo_server_thread_start(_polyOscThread);
	lo_server_thread_start(_monomeOscThread);
	char initMessage[1];
	initMessage[0]= GETMODE;
	_polySerial->write(initMessage, 1);
	
	cout<<"Polynome 5K initialized and starting...\n";
}

void polynome::setMode(int m) {
	_mode = m;
}

int polynome::getMode() {
	return _mode;
}

void polynome::clearColorMatrix() {
	for (int i=0; i<8; i++) {
		for (int j=0; j<8; j++) {
			colorMatrix[i][j][0] = 0;
			colorMatrix[i][j][1] = 0;
			colorMatrix[i][j][2] = 0;
		}
	}
}

void polynome::sendColorMatrix() {
	char m[3];
	for (int i=0; i<8; i++) {
		for (int j=0; j<8; j++) {
			m[0] = RGBLED + i;
			m[1] = (j<<4) + colorMatrix[i][j][0];
			m[2] = (colorMatrix[i][j][1]<<4) + colorMatrix[i][j][2];
			_polySerial->write(m, 3);
		}
		usleep(5000);
	}
}

/*****************OSC MESSAGE HANDLERS***************/
/****************************************************/
void polyOscError(int num, const char *msg, const char *path)
{
    printf("polynome server error %d in path %s: %s\n", num, path, msg);
    fflush(stdout);
}



//***********************************//
//Monome Specific Event handlers
//***********************************//
int m_LedHandler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data)
{
	polynome *SELF = (polynome *)user_data;
	char message[2];
	int x = argv[0]->i;
	int y = argv[1]->i;
	message[0] = ((argv[2]->i)+2)<<4;
	message[1] = (x<<4)|y;
	if (x<8 && y<8)
		SELF->getSerialDevice()->write(message, 2);
	fflush(stdout);
	return 0;
}

int m_ClearHandler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data)
{
	polynome *SELF = (polynome *)user_data;
	char message[1];
	message[0] = M_CLEAR + argv[0]->i;
	SELF->getSerialDevice()->write(message, 1);
	fflush(stdout);
	return 0;
}

int m_LedColHandler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data)
{
	polynome *SELF = (polynome *)user_data;
	int col = argv[0]->i;
	if (col<8) {	
		char message[2];
		message[0] = M_LEDCOL1 + col;
		message[1] = argv[1]->i;
		SELF->getSerialDevice()->write(message, 2);
	}
	else {
		cout<<"POL:OSC:COL col index out of range"<<endl;
	}
	fflush(stdout);
	return 0;
}

int m_LedRowHandler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data)
{
	polynome *SELF = (polynome *)user_data;
	int row = argv[0]->i;
	if (row<8) {
		char message[2];
		message[0] = M_LEDROW1 + row;
		message[1] = argv[1]->i;
		SELF->getSerialDevice()->write(message, 2);
	}
	else {
		cout<<"OSC:MON:ROW row index out of range"<<endl;
	}
	fflush(stdout);
	return 0;
}

int m_frameHandler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data)
{
	polynome *SELF = (polynome *)user_data;
	char message[9];
	message[0] = M_LEDFRAME;
	for (int i=0; i<8; i++)
		message[i+1] = argv[i]->i;
	SELF->getSerialDevice()->write(message, 9);
	fflush(stdout);
	return 0;
}

int m_SysPrefixHandler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data)
{
	polynome *SELF = (polynome *)user_data;
	string newPrefix = (&argv[argc-1]->s);
	if ( SELF->getPrefix() != newPrefix ) {
		
		lo_server_thread_add_method(SELF->getMonomeServerThread(), (newPrefix+"/led").c_str(), "iii", m_LedHandler, SELF);
		lo_server_thread_add_method(SELF->getMonomeServerThread(), (newPrefix+"/clear").c_str(), "i", m_ClearHandler, SELF);
		lo_server_thread_add_method(SELF->getMonomeServerThread(), (newPrefix+"/led_col").c_str(), "ii", m_LedColHandler, SELF);
		lo_server_thread_add_method(SELF->getMonomeServerThread(), (newPrefix+"/led_row").c_str(), "ii", m_LedRowHandler, SELF);
		lo_server_thread_add_method(SELF->getMonomeServerThread(), (newPrefix+"/frame").c_str(), "iiiiiiiiii", m_frameHandler, SELF);
		
		lo_server_thread_del_method(SELF->getMonomeServerThread(), (SELF->getPrefix()+"/led").c_str(), "iii");
		lo_server_thread_del_method(SELF->getMonomeServerThread(), (SELF->getPrefix()+"/clear").c_str(), "i");
		lo_server_thread_del_method(SELF->getMonomeServerThread(), (SELF->getPrefix()+"/led_col").c_str(), "ii");
		lo_server_thread_del_method(SELF->getMonomeServerThread(), (SELF->getPrefix()+"/led_row").c_str(), "ii");
		lo_server_thread_del_method(SELF->getMonomeServerThread(), (SELF->getPrefix()+"/frame").c_str(), "iiiiiiiii");
		
		char m[1] = {M_CLEAR};
		SELF->getSerialDevice()->write(m, 1);		//Clear monome, some devices don't do this automatically, they should perhaps...
		SELF->setPrefix(newPrefix);
		lo_send(SELF->getMonomeSender(),"/sys/prefix","is",0,(SELF->getPrefix()).c_str());
		cout<<"OSC:PREFIX: new prefix is: "<<SELF->getPrefix()<<endl;
	}
	else {
		lo_send(SELF->getMonomeSender(),"/sys/prefix","is",0,(SELF->getPrefix()).c_str());
		cout<<"OSC:PREFIX: same prefix! no changes made."<<endl;
	}
	return 0;
}

int m_SysTestHandler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data)
{
	polynome *SELF = (polynome *)user_data;
	char message[1];
	message[0] = M_CLEAR + argv[0]->i;
	SELF->getSerialDevice()->write(message, 1);
	fflush(stdout);
	return 0;
}

int m_SysReportHandler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data)
{
	polynome *SELF = (polynome *)user_data;
	lo_send(SELF->getMonomeSender(),"/sys/devices","i",1);
	lo_send(SELF->getMonomeSender(),"/sys/prefix","is",0,(SELF->getPrefix()).c_str());
	lo_send(SELF->getMonomeSender(),"/sys/type","is",0,"64");
	lo_send(SELF->getMonomeSender(),"/sys/cable","is",0,"left");
	lo_send(SELF->getMonomeSender(),"/sys/offset","iii",0,0,0);
	return 0;
}




//***********************************//
//Polynome Specific Event Handlers
//***********************************//
int p_ModeHandler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data)
{
	polynome *SELF = (polynome *)user_data;
	char message[1];
	message[0] = MODE + (argv[0]->i);
	//printf("OSC:POLY:MODE: Sending to serial: 0x%u\n",message[0]);
	SELF->getSerialDevice()->write(message, 1);
	fflush(stdout);
	return 0;
}

//Message format is /polynome/monocolor r g b (r,g,b values go from 0->3, 4 bits)
int p_MonoColorHandler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data)
{
	polynome *SELF = (polynome *)user_data;
	char message[2];
	message[0] = MONOCOLOR + argv[0]->i;
	message[1] = ((argv[1]->i)<<4)+(argv[2]->i);
	SELF->getSerialDevice()->write(message, 2);
	fflush(stdout);
	return 0;
}

//Message format is /polynome/led x y r g b
int p_LedHandler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data)
{
	polynome *SELF = (polynome *)user_data;
	int x = (argv[0]->i);
	int y = (argv[1]->i);
	int red = (argv[2]->i);
	int grn = (argv[3]->i);
	int blu = (argv[4]->i);
	char message[3];
	message[0] = RGBLED + x;
	message[1] = (y<<4) + red;
	message[2] = (grn<<4) + blu;
	SELF->getSerialDevice()->write(message, 3);
	usleep(5000);
	fflush(stdout);
	return 0;
}

//Message format is /polynome/row Row r g b
int p_RowHandler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data)
{
	polynome *SELF = (polynome *)user_data;
	int y = (argv[0]->i);
	int red = (argv[1]->i);
	int grn = (argv[2]->i);
	int blu = (argv[3]->i);
	char message[3];
	for (int i=0; i<8; i++) {
		message[0] = RGBLED + i;
		message[1] = (y<<4) + red;
		message[2] = (grn<<4) + blu;
		SELF->getSerialDevice()->write(message, 3);
	}
	usleep(10000);
	fflush(stdout);
	return 0;
}

//Message format is /polynome/col Col r g b
int p_ColHandler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data)
{
	polynome *SELF = (polynome *)user_data;
	int x = (argv[0]->i);
	int red = (argv[1]->i);
	int grn = (argv[2]->i);
	int blu = (argv[3]->i);
	char message[3];
	for (int i=0; i<8; i++) {
		message[0] = RGBLED + x;
		message[1] = (i<<4) + red;
		message[2] = (grn<<4) + blu;
		SELF->getSerialDevice()->write(message, 3);
	}
	usleep(10000);
	return 0;
}


int p_ColorDumpHandler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data)
{
	polynome *SELF = (polynome *)user_data;
	string command = &argv[0]->s;
	if( command=="start" ) {
		SELF->clearColorMatrix();
	}
	if (command=="/polynome/led") {
		int x = argv[1]->i;
		int y = argv[2]->i;
		int red = (argv[3]->i);
		int grn = (argv[4]->i);
		int blu = (argv[5]->i);
		SELF->colorMatrix[x][y][0] = red;
		SELF->colorMatrix[x][y][1] = grn;
		SELF->colorMatrix[x][y][2] = blu;
	}
	if (command=="/polynome/row") {
		int row = argv[1]->i;
		int red = (argv[2]->i);
		int grn = (argv[3]->i);
		int blu = (argv[4]->i);
		for(int i=0;i<8;i++)
		{
			SELF->colorMatrix[i][row][0] = red;
			SELF->colorMatrix[i][row][1] = grn;
			SELF->colorMatrix[i][row][2] = blu;
		}
	}
	if (command=="/polynome/col") {
		int col = argv[1]->i;
		int red = (argv[2]->i);
		int grn = (argv[3]->i);
		int blu = (argv[4]->i);
		for(int i=0;i<8;i++)
		{
			SELF->colorMatrix[col][i][0] = red;
			SELF->colorMatrix[col][i][1] = grn;
			SELF->colorMatrix[col][i][2] = blu;
		}
	}
	if( command=="end" ) {
		SELF->sendColorMatrix();	
	}
	fflush(stdout);
	return 0;
}

//Message format is /polynome/arcade led state (i.e. /polynome/arcade 3 0, sets button 3 led to off)
int p_ArcadeHandler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data)
{	
	polynome *SELF = (polynome *)user_data;
	vector<unsigned char> midimessage;
	char message[1];
	message[0] = (((argv[1]->i)<<4)+ALEDOFF) + (argv[0]->i);
	SELF->getSerialDevice()->write(message, 1);
	if( (argv[0]->i)<3) {
		midimessage.push_back(PMIDI_CC_MESSAGE);
		midimessage.push_back( (argv[0]->i)+PMIDI_ARCADECC);
		midimessage.push_back(127*(argv[1]->i));
		SELF->getMidiOut()->sendMessage( &midimessage );
	}
	fflush(stdout);
	return 0;
}

int p_NixieHandler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data)
{
	
	polynome *SELF = (polynome *)user_data;
	if(SELF->nixieUpdate()==true) {
		char message[4];
		int bcd = 0;
		float t=0.0;
		message[0] = NIXIESET;
		switch( types[0] )
		{
			case 'i':
				bcd = intToBCD(argv[0]->i);
				message[1] = bcd>>8;
				message[2] =  ((bcd&0xFF)>10?(bcd&0xFF):((bcd&0xFF)|0xF0));
				message[3] = NIXIEDEC + 2;
				break;
			case 'f':
				t = argv[0]->f;
				if (t>=1000) {
					bcd = intToBCD((int)t);
					message[1] = bcd>>8;
					message[2] = bcd & 0x00FF;
					message[3] = NIXIEDEC + 1;
				}
				if (t>=100 && t<1000) {
					bcd = intToBCD((int)(t*10.1));
					message[1] = bcd>>8;
					message[2] = bcd & 0x00FF;
					message[3] = NIXIEDEC + 1;
				}
				if (t>=10 && t<100) {
					bcd = intToBCD(t*100);
					message[1] = bcd>>8;
					message[2] = ((bcd&0xFF)>10?(bcd&0xFF):((bcd&0xFF)|0xF0));
					message[3] = NIXIEDEC + 2;
				}
				if (t>=0 && t<10) {
					bcd = intToBCD(t*1000);
					message[1] = bcd>>8;
					message[2] = ((bcd&0xFF)>10?(bcd&0xFF):((bcd&0xFF)|0xF0));
					message[3] = NIXIEDEC + 2;
				}
				break;
			default:
				break;
		}
		SELF->getSerialDevice()->write(message, 4);

	}
	fflush(stdout);
	return 0;
}

int p_UpdateHandler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data)
{
	//polynome *SELF = (polynome *)user_data;
	
	fflush(stdout);
	return 0;
}

int polynome::keypadAnalyze(char key)
{
	int temp=0;
	_nixieBCD = 0;
	
	if(key != '*' && key !='#')
	{
		_nixieDisplay.append(1,key);
		_nixieDisplay.erase(0,1);
		if(_nixieDisplay.length()>4)
			_nixieDisplay.erase(4);
		_nixieUpdate = false;
	}
	if(key == '*')
	{
		_nixieDisplay.insert(0, 1, '0');
		_nixieDisplay.erase(4);
		if(_nixieDisplay =="0000")
			_nixieUpdate = false;
	}
	if(key == '#')
	{
		if(_nixieDisplay=="0000") {
			char m[1];
			m[0] = MODEMENU;
			_polySerial->write(m,1);
		}
		else {
			lo_send(_polynomeOscSender,"/polynome/keypad","i",atoi(_nixieDisplay.c_str()));
			_nixieDisplay = "0000";
		}
		_nixieUpdate = true;
		
	}
	for(int i=0; i<4; i++)
	{
		temp = ((char)(_nixieDisplay[3-i])-48)<<(4*i);
		_nixieBCD += temp;
	}
	return _nixieBCD;
	
}

/* itoa:  convert n to characters in s */
uint8* myITOA(int n)
{
	uint8 b[4];
	b[0] = (n/1000)+48;
	b[1] = ((n%1000)/100)+48;
	b[2] = ((n%100)/10)+48;
	b[3] = (n%10)+48;
	return b;
} 


int intToBCD(int n)
{
	int thisBCD = n/1000;
	thisBCD = thisBCD<<4;
	thisBCD += (n%1000)/100;
	thisBCD = thisBCD<<4;
	thisBCD += (n%100)/10;
	thisBCD = thisBCD<<4;
	thisBCD += n%10;
	return thisBCD;
}


