/*
 *  relayBox.h
 *  polynomeSerial
 *
 *  Created by Colin Mann on 6/25/10.
 *  Copyright 2010 Mannchester Analog. All rights reserved.
 *
 */
#ifndef _RELAY_BOX
#define _RELAY_BOX

#include "lo/lo.h"
#include "types.h"
#include "AsynchronousSerialDeviceReader.h"
#include "SerialDevice.h"
#include <string>
#include "RtMidi.h"


#define RELAYOFF				0x00				//0x0c, c = relay to set	
#define RELAYON					0x10				//0x0c, c = relay to turn off
#define RB_SETALL				0x20				//0x20 0bSSSSSSSS, state of each relay, two byte message.
#define DEBUG					0x70	
	
#define SWITCHON				0x40				//0x00 0x0c, c = channel switch pressed
#define SWITCHOFF				0x50				//0x0c 0x0c, c = channel switch released

#ifndef UART_ERRORS
#define UART_ERRORS
#define UART_FRAME_ERROR		0x08              /* Framing Error by UART       */
#define UART_OVERRUN_ERROR		0x04              /* Overrun condition by UART   */
#define UART_BUFFER_OVERFLOW	0x02              /* receive ringbuffer overflow */
#define UART_NO_DATA			0x01              /* no receive data available   */
#endif


#define messageGetType(message)			((message).data0 & 0xF0) 

using namespace std;

void relayOscError(int num, const char *msg, const char *path);
int relayChannelHandler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data);
int relayHandler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data);

#define RELAY_LISTEN_PORT		"10005"
#define RELAY_PORT_SENDER		"10004"

#define RMIDI_CC_MESSAGE		176
#define RMIDI_CC_START			70

#define NOTEON					144
#define NOTEOFF					128
#define RMIDI_START				60		//First channel midi note, for received data (middle C)

typedef struct {
	uint8 data0;
	uint8 data1;
	
} r_message;

class relay{
	
public:
	
	relay();
	virtual ~relay();
	
	void start();

	lo_server_thread getServerThread(void) { return _relayOscThread; }
	lo_address getSender(void) { return _relayOscSender; }
	RtMidiOut * getMidiOut(void) { return _relayMidiOut; }	
	SerialDevice * getSerialDevice(void) { return _relaySerial; }
	
	
private:
	
	lo_server_thread _relayOscThread;
	lo_address _relayOscSender;

	AsynchronousSerialDeviceReader _relayDeviceReader;
	SerialDevice *_relaySerial;
	
	RtMidiOut *_relayMidiOut;
	RtMidiIn *_relayMidiIn;
	
};

#endif