/*
 *  voltageController.cpp
 *  polynomeSerial
 *
 *  Created by Colin Mann on 6/26/10.
 *  Copyright 2010 Mannchester Analog. All rights reserved.
 *
 */

#include <stdio.h>
#include <iostream>
#include <stdint.h>
#include "lo/lo.h"
#include "SerialDevice.h"
#include "voltageController.h"

using namespace std;

static int _VoltageController_SerialDeviceMessageReceivedCallback(SerialDevice *device, char *data, size_t len, void *userData) {
	
	voltageController *SELF = (voltageController *)userData;
	vector<unsigned char> midimessage(3,0);
	v_message *message = (v_message *)data;
	char out[1];
	switch (messageGetType(*message)) {
		case REPLUS:
			midimessage[0] = (VC_MIDI_CC_MESSAGE);
			midimessage[1] = ( (message->data1)+VC_MIDI_ENCODER_CC);
			midimessage[2] = (VC_MIDI_INC_MESSAGE(SELF->coarseAdjust[message->data1]));
			SELF->getMidiOut()->sendMessage( &midimessage );
			SELF->displayValues[message->data1]+= 1+(COARSEVALUE-1)*(SELF->coarseAdjust[message->data1]);
			SELF->updateDisplayNumber(SELF->displayValues[message->data1],message->data1);
			lo_send(SELF->getSender(),"/voltageController/encoder","ii",message->data1,1);
			break;
		case REMINUS:
			midimessage[0] = (VC_MIDI_CC_MESSAGE);
			midimessage[1] = ( (message->data1)+VC_MIDI_ENCODER_CC);
			midimessage[2] = (VC_MIDI_DEC_MESSAGE(SELF->coarseAdjust[message->data1]));
			SELF->getMidiOut()->sendMessage( &midimessage );
			SELF->displayValues[message->data1]-= 1+(COARSEVALUE-1)*(SELF->coarseAdjust[message->data1]);
			SELF->updateDisplayNumber(SELF->displayValues[message->data1],message->data1);
			lo_send(SELF->getSender(),"/voltageController/encoder","ii",message->data1,-1);
			break;
		case PBDOWN:
			//midimessage.push_back(VC_MIDI_CC_MESSAGE);
			//midimessage.push_back( (message->data1)+VC_MIDI_PUSHB_CC);
			//midimessage.push_back(127);
			//SELF->getMidiOut()->sendMessage( &midimessage );
			SELF->coarseAdjust[message->data1] ^= 1;
			out[0] = ((SELF->coarseAdjust[message->data1])<<4) + ((message->data1)<<2)+3;
			SELF->getSerialDevice()->write(out, 1);
			lo_send(SELF->getSender(),"/voltageController/coarse","ii",message->data1,SELF->coarseAdjust[message->data1]);
			break;
		case PBUP:
			//midimessage.push_back(VC_MIDI_CC_MESSAGE);
			//midimessage.push_back( (message->data1)+VC_MIDI_PUSHB_CC);
			//midimessage.push_back(0);
			//SELF->getMidiOut()->sendMessage( &midimessage );
			//lo_send(SELF->getSender(),"/voltageController/button","ii",message->data1,0);
			break;
		case TOGUP:
			midimessage[0] = (VC_MIDI_CC_MESSAGE);
			midimessage[1] = ( (message->data1)+VC_MIDI_TOGGLE_CC);
			midimessage[2] = 127;
			SELF->getMidiOut()->sendMessage( &midimessage );
			//lo_send(SELF->getSender(),"/voltageController/toggle","ii",message->data1,1);
			break;
		case TOGDOWN:
			midimessage[0] = (VC_MIDI_CC_MESSAGE);
			midimessage[1] = ( (message->data1)+VC_MIDI_TOGGLE_CC);
			midimessage[2] = 0;
			SELF->getMidiOut()->sendMessage( &midimessage );
			//lo_send(SELF->getSender(),"/voltageController/toggle","ii",message->data1,0);
			break;
		case SWITCHES:
			//set my switch values
			break;
		case STATUS:
			//receive status ok message, more of a query than anything
			break;
		case DEBUG:
			switch (message->data1) {
				case UART_FRAME_ERROR:
					printf("VOLTAGECONTROLLER:DEBUG:A framing error has occured on the UART bus.\n");
					break;
				case UART_OVERRUN_ERROR:
					printf("VOLTAGECONTROLLER:DEBUG:A buffer overrun error has occured, sending data too fast.\n");
					break;
				case UART_BUFFER_OVERFLOW:
					printf("VOLTAGECONTROLLER:DEBUG:A buffer overflow error has occured, sending data too fast.\n");
					break;
				default:
					printf("VOLTAGECONTROLLER:DEBUG:DEFAULT: byte 0x%X\n",message->data1);
			}
			break;
		default:
			printf("VOLTAGECONTROLLER:ERROR:OUT OF SYNC!!\nVOLTAGECONTROLLER:DEFAULT:Byte 0 is: 0x%X, Byte 1 is: 0x%X\n",(message->data0),message->data1);
			return -1;
			break;
	}
	return 0;
}

void voltageControllerMidiCallback( double deltatime, vector< unsigned char > *message, void *userData )
{
	voltageController *SELF = (voltageController *)userData;
	unsigned int nBytes = message->size();

	if(nBytes>2)
	{
		if((message->at(0))==VC_MIDI_CC_MESSAGE)
		{
			int channel = (message->at(1))-VC_MIDI_ENCODER_CC;
			int value = message->at(2);
			SELF->displayValues[channel] = value;
			char buffer[5];
			if( channel < 4 )
			{
				buffer[0] = SETCHANNEL+channel;
				sprintf(buffer+1,"%04d",value);
				SELF->getSerialDevice()->write(buffer, 5);
			}
			else {
				printf("Other midi data from channel %i, value: %i\n",channel,value);
			}
		}
	}

			
	
}

voltageController::voltageController() {
	
	for(int i=0;i<4;i++) {
		displayValues[i] = 0;
		coarseAdjust[i] = 0;
	}
	
}

voltageController::~voltageController() {
	_voltageControllerDeviceReader.stopReading();
	_voltageControllerDeviceReader.removeSerialDevice(_voltageControllerSerial);
	_voltageControllerMidiOut->closePort();
	delete _voltageControllerMidiOut;
	delete _voltageControllerMidiIn;
}

void voltageController::start() {
	
	_voltageControllerSerial = new SerialDevice("/dev/tty.usbserial-voltcont",B115200);
	_voltageControllerDeviceReader.addSerialDevice(_voltageControllerSerial, 2, _VoltageController_SerialDeviceMessageReceivedCallback, this);
	_voltageControllerSerial->flush();
	
	_voltageControllerMidiOut = new RtMidiOut();
	_voltageControllerMidiOut->openVirtualPort("Voltage Controller");
	
	_voltageControllerMidiIn = new RtMidiIn();
	_voltageControllerMidiIn->openVirtualPort("Voltage Controller");
	_voltageControllerMidiIn->setCallback(&voltageControllerMidiCallback, this);
	_voltageControllerMidiIn->ignoreTypes(false, false, false);
	
	//Create OSC thread
	_voltageControllerOscThread = lo_server_thread_new(VC_LISTEN_PORT, voltageControllerOscError);
	//Create OSC message senders
	_voltageControllerOscSender = lo_address_new(NULL, VC_PORT_SENDER);	//Null address is localmachine
	
	
	//Voltage Controller messages
	lo_server_thread_add_method(_voltageControllerOscThread, ("/voltageController/lcd"), "ii", voltageControllerNumberHandler, this);
	lo_server_thread_add_method(_voltageControllerOscThread, ("/voltageController/lcd"), "is", voltageControllerAsciiHandler, this);
	lo_server_thread_add_method(_voltageControllerOscThread, ("/voltageController/coarse"), "ii", voltageControllerCoarseHandler, this);
	//Messages passed to racMini LCD (out UART1 on vc)
	lo_server_thread_add_method(_voltageControllerOscThread, ("/lcd"), "i", lcdStateHandler, this);
	lo_server_thread_add_method(_voltageControllerOscThread, ("/lcd/text"), "s", lcdStringHandler, this);
	lo_server_thread_add_method(_voltageControllerOscThread, ("/lcd/baud"), "i", lcdBaudHandler, this);
	lo_server_thread_add_method(_voltageControllerOscThread, ("/lcd/reset"), "", lcdResetHandler, this);
	
	
	_voltageControllerDeviceReader.startReading();
	lo_server_thread_start(_voltageControllerOscThread);
	
	//Set all decimals to off, for consistency of coarse setting
	char m[4] = {3,7,11,15};
	getSerialDevice()->write(m, 4);
	m[0] = LCDOFF;
	getSerialDevice()->write(m, 1);
	
	
	cout<<"Voltage Controller initialized and starting...\n";
}

void voltageController::updateDisplayNumber(int n,int c)
{
	if(n>127) {
		displayValues[c] = 127;
		return;
	}
	if(n<0) {
		displayValues[c] = 0;
		return;
	}		
	char buffer[5];
	if( c < 4 )
	{
		buffer[0] = SETCHANNEL+c;
		sprintf(buffer+1,"%04d",n);
		getSerialDevice()->write(buffer, 5);
	}
}

/******************VC MESSAGE HANDLERS***************/
/****************************************************/
void voltageControllerOscError(int num, const char *msg, const char *path)
{
    printf("voltageController server error %d in path %s: %s\n", num, path, msg);
    fflush(stdout);
}


int voltageControllerNumberHandler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data)
{
	voltageController *SELF = (voltageController *)user_data;
	char buffer[5];
	int channel = argv[0]->i;
	int value = argv[1]->i;
	value = value%9999;
	buffer[0] = SETCHANNEL+channel;
	sprintf(buffer+1,"%4d",value);
	printf("The buffer is: %s\n",buffer);
	SELF->getSerialDevice()->write(buffer, 5);
	fflush(stdout);
	return 0;
}

int voltageControllerAsciiHandler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data)
{
	voltageController *SELF = (voltageController *)user_data;
	char buffer[5];
	int channel = argv[0]->i;
	string text = &argv[1]->s;
	text.insert(0, 1, SETCHANNEL+channel);
	printf("The buffer is: %s\n",text.c_str());
	SELF->getSerialDevice()->write((char *)(text.substr(0, 5)).c_str(), 5);
	fflush(stdout);
	return 0;
}


int voltageControllerCoarseHandler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data)
{
	voltageController *SELF = (voltageController *)user_data;
	int channel = argv[0]->i;
	int state = argv[1]->i;
	char out[1];
	SELF->coarseAdjust[channel] = state;
	out[0] = ((SELF->coarseAdjust[channel])<<4) + ((channel)<<2)+3;
	SELF->getSerialDevice()->write(out, 1);
	fflush(stdout);
	return 0;
}


//***********************************//
//LCD Specific Event Handlers
//***********************************//
int lcdStateHandler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data)
{
	voltageController *SELF = (voltageController *)user_data;
	char message[1];
	message[0] = LCDOFF + ((argv[0]->i)>0?1:0);
	SELF->getSerialDevice()->write(message, 1);
	cout<<"got to reset handler"<<endl;
	fflush(stdout);
	return 0;
}
int lcdStringHandler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data)
{
	voltageController *SELF = (voltageController *)user_data;
	string s = (&argv[0]->s);
	s.insert(s.begin(),1, LCDSTART);
	s.append(1,EOT);
	SELF->getSerialDevice()->write((char *)s.c_str(), s.size());
	fflush(stdout);
	return 0;
}

int lcdBaudHandler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data)
{
	voltageController *SELF = (voltageController *)user_data;
	int baud = 16;			//<control> P
	switch (argv[0]->i) {
		case 2400:
			baud = 11;		//<control> K
			break;
		case 4800:
			baud = 12;		//<control> L
			break;
		case 9600:
			baud = 13;		//<control> M
			break;
		case 14400:
			baud = 14;		//<control> N
			break;
		case 19200:
			baud = 15;		//<control> O
			break;
		case 38400:
			baud = 16;		//<control> P
			break;
		default:
			baud = 16;		//<control> P
			break;
	}
	char message[4];
	message[0] = LCDSTART;
	message[1] = 124;
	message[2] = baud;
	message[3] = EOT;
	SELF->getSerialDevice()->write((char *)message,4);
	fflush(stdout);
	return 0;
}

int lcdResetHandler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data)
{
	voltageController *SELF = (voltageController *)user_data;
	char message[1];
	message[0] = LCDRESET;
	SELF->getSerialDevice()->write(message, 1);
	fflush(stdout);
	return 0;
}
