/*
 *  voltageController.h
 *  polynomeSerial
 *
 *  Created by Colin Mann on 6/26/10.
 *  Copyright 2010 Mannchester Analog. All rights reserved.
 *
 */

#ifndef _VOLTAGE_CONTROLLER
#define _VOLTAGE_CONTROLLER

#include "lo/lo.h"
#include "types.h"
#include "AsynchronousSerialDeviceReader.h"
#include "SerialDevice.h"
#include <string>
#include "RtMidi.h"


/***********FROM COMPUTER TO DEVICE MESSAGES************/
//voltage controller commands... for messages.h
//Split commands (command includes some data)
#define	DPOINTOFF	0x00		//Lower 4 Bits represent 0x0ccdd (cc = channel, 0-3; dd = digit,0-3)
#define DPOINTON	0x10		//Lower 4 Bits represent 0x1ccdd (cc = channel, 0-3; dd = digit,0-3)
#define SETCHANNEL	0x20		//Lower 4 Bits represent channel(color) to set, message is followed by 4 bytes\

//specific commands (no data in command)
#define INIT		0x70		//Run initialization sequence
#define SETALL		0x71		//Command is followed by 16 bytes, in order from left to right, setting all values
#define SW_QUERY	0x72		//Asks for current switch configuration, also run on startup, returns 1 byte 8 bitss
#define DEVICEOK	0x73

#define LCDOFF		0x7C		//Turn on lcd power transistor
#define LCDON		0x7D		//Turn off lcd power transistor
#define LCDRESET	0x7E		//On/oFF/On lcd power transistor

#define LCDSTART	0x60		
#define EOT			0x04

/***********FROM DEVICE TO COMPUTER MESSAGES************/
#define REPLUS		0x10
#define REMINUS		0x20
#define PBDOWN		0x30		//Lower 4 bits are the push button values (buttons 0-4)
#define PBUP		0x40		//This byte is followed by another byte with all switch value data in it.
#define TOGDOWN		0x50
#define TOGUP		0x60
#define SWITCHES	0xE0
#define STATUS		0xF0		//Byte is returned when status? byte is sent, to see if device is responding correctly.
#define DEBUG		0x70

#ifndef UART_ERRORS
#define UART_ERRORS
#define UART_FRAME_ERROR		0x08              /* Framing Error by UART       */
#define UART_OVERRUN_ERROR		0x04              /* Overrun condition by UART   */
#define UART_BUFFER_OVERFLOW	0x02              /* receive ringbuffer overflow */
#define UART_NO_DATA			0x01              /* no receive data available   */
#endif

#define messageGetType(message)			((message).data0 & 0xF0) 

using namespace std;

void voltageControllerOscError(int num, const char *msg, const char *path);
int voltageControllerNumberHandler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data);
int voltageControllerAsciiHandler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data);
int voltageControllerCoarseHandler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data);

int lcdStateHandler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data);
int lcdStringHandler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data);
int lcdBaudHandler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data);
int lcdResetHandler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data);

#define VC_LISTEN_PORT		"10003"
#define VC_PORT_SENDER		"10002"

#define NOTEON					144
#define NOTEOFF					128
#define RMIDI_START				60		//First channel midi note, for received data (middle C)

#define VC_MIDI_CC_MESSAGE		176
#define VC_MIDI_ENCODER_CC		16
#define VC_MIDI_PUSHB_CC		20
#define VC_MIDI_TOGGLE_CC		24
#define COARSEVALUE				4
#define VC_MIDI_INC_MESSAGE(coarse)		(66 + (coarse*COARSEVALUE))
#define VC_MIDI_DEC_MESSAGE(coarse)		(62 - (coarse*COARSEVALUE))

typedef struct {
	uint8 data0;
	uint8 data1;
	
} v_message;

class voltageController{
	
public:
	
	voltageController();
	virtual ~voltageController();
	
	void start();
	
	lo_server_thread getServerThread(void) { return _voltageControllerOscThread; }
	lo_address getSender(void) { return _voltageControllerOscSender; }
	RtMidiOut * getMidiOut(void) { return _voltageControllerMidiOut; }	
	SerialDevice * getSerialDevice(void) { return _voltageControllerSerial; }
	
	void updateDisplayNumber(int number,int channel);
	
	int	displayValues[4];
	int coarseAdjust[4];
	
private:
	
	lo_server_thread _voltageControllerOscThread;
	lo_address _voltageControllerOscSender;
	
	AsynchronousSerialDeviceReader _voltageControllerDeviceReader;
	SerialDevice *_voltageControllerSerial;
	
	RtMidiOut *_voltageControllerMidiOut;
	RtMidiIn *_voltageControllerMidiIn;
	
};

#endif